
#ifndef __MUSITORIUS_CONFIG_FILE_H__
#define __MUSITORIUS_CONFIG_FILE_H__

//Taken from VoiceMan at 2011-01-06;

#define MUSITORIUS_BEGIN_PARAM_TABLE(cl) int cl::params(const std::string& section, const std::string& param) const {
#define MUSITORIUS_RELAX_SECTION(s) if (section==(s)) return AnyValue
#define MUSITORIUS_DECLARE_PARAM(s,p) if (section==(s) && param==(p)) return AnyValue
#define MUSITORIUS_DECLARE_BOOLEAN_PARAM(s,p) if (section==(s) && param==(p)) return BooleanValue
#define MUSITORIUS_DECLARE_STRING_PARAM(s,p) if (section==(s) && param==(p)) return StringValue
#define MUSITORIUS_DECLARE_INT_PARAM(s,p) if (section==(s) && param==(p)) return IntValue
#define MUSITORIUS_DECLARE_UINT_PARAM(s,p) if (section==(s) && param==(p)) return UintValue
#define MUSITORIUS_DECLARE_DOUBLE_PARAM(s,p) if (section==(s) && param==(p)) return DoubleValue
#define MUSITORIUS_END_PARAM_TABLE return InvalidValue;}

class ConfigFileException: public MusitoriusException
{
public:
  enum {
    InvalidSectionHeader = 1,
    EmptySectionName = 2,
    ParameterWithoutSection = 3,
    InvalidUnquotedValue = 4,
    InvalidLineBeginning = 5,
    InvalidParameterName = 6,
    InvalidCharAfterEquals = 7,
    UnexpectedLineEnd = 8,
    DublicatedKey = 9
  };

  ConfigFileException(int code)
    : m_code(code), m_line(0) {}

  ConfigFileException(int code, const std::string& fileName, size_t line)
    : m_code(code), m_fileName(fileName), m_line(line) 
  {
    assert(m_line > 0);
  }

  virtual ~ConfigFileException() {}

  int getCode() const 
  {
    return m_code;
  }

  size_t getLine() const
  {
    assert(m_line > 0);
    return m_line;
  }

  std::string getFileName() const
  {
    return m_fileName;
  }

  std::string getDescr() const
  {
    switch(m_code)
      {
      case InvalidSectionHeader:
	return "invalid section header format";
      case EmptySectionName:
	return "missed section name";
      case ParameterWithoutSection:
	return "config file section is not opened";
      case InvalidUnquotedValue:
	return "invalid character in value without quotas";
      case InvalidLineBeginning:
	return "invalid character before parameter name";
      case InvalidParameterName:
	return "invalid character in parameter name";
      case InvalidCharAfterEquals:
	return "invalid character after \'=\' sign";
      case UnexpectedLineEnd:
	return "unexpected line end or line is incomplete";
      case DublicatedKey:
	return "dublicated parameter key in section";
      default:
	assert(0);
      } //switch(m_code);
    return "";
  }

  std::string getMessage() const 
  {
    if (m_line == 0)
      return getDescr();
    std::ostringstream s;
    s << m_fileName << "(" << m_line << "):" << getDescr();
    return s.str();
  }

  void makeLogReport(int level) const
  {
    logMsg(level, "config file error:%s", getMessage().c_str());
  }

private:
  const int m_code;
  const std::string m_fileName;
  const size_t m_line;
}; // class ConfigFileException;

class ConfigFileValueTypeException: public MusitoriusException
{
public:
  ConfigFileValueTypeException(const std::string& descr)
    : m_descr(descr) {}

  virtual ~ConfigFileValueTypeException() {}

  std::string getMessage() const
  {
    return m_descr;
  }

  void makeLogReport(int level) const
  {
    logMsg(level, "config file value type mismatch:%s", getMessage().c_str());
  }

private:
  const std::string m_descr;
}; //class ConfigFileValueTypeException;

class ConfigFileSection: public std::map<std::string, std::string>
{
public:
  ConfigFileSection() {}

  ConfigFileSection(const std::string& name)
    : m_name(name) {}

  virtual ~ConfigFileSection() {}

  std::string getName() const 
  {
    return m_name;
  }

  void setName(const std::string& name)
  {
    m_name=name;
  }

  bool has(const std::string& name) const 
  {
    return find(name) != end();
  }

  void add(const std::string& param, const std::string& value)
  {
    insert(value_type(param, value));
  }

  std::string operator [](const std::string& name) const
  {
    const_iterator it=find(name);
    assert(it!=end());
    if (it == end())
      return "";
    return it->second;
  }

private:
  std::string m_name;
}; // class ConfigFileSection;

typedef std::vector<ConfigFileSection> ConfigFileSectionVector;
typedef std::list<ConfigFileSection> ConfigFileSectionList;

class ConfigFile
{
public:
  ConfigFile() {}

  virtual ~ConfigFile() {}

  void load(const std::string& fileName);
  void checkParams() const;
  ConfigFileSectionVector::size_type getSectionCount() const;
  const ConfigFileSection& getSection(ConfigFileSectionVector::size_type index) const;
  const ConfigFileSection& findSection(const std::string& name) const;
  bool hasSection(const std::string& name) const;

protected:
  enum {AnyValue = -1, InvalidValue = 0, BooleanValue = 1, StringValue = 2, IntValue = 3, UintValue = 4, DoubleValue = 5};
  virtual int params(const std::string& section, const std::string& param) const = 0;

private:
  int process(const std::string& line, int& code, std::string& str1, std::string& str2) const;
  void checkParamsInSection(const ConfigFileSection& section) const;

private:
  ConfigFileSectionVector m_sections;
}; //class ConfigFile;

#endif // __MUSITORIUS_CONFIG_FILE_H__
