
#include"musitorius.h"
#include"MPlayer.h"

#define SHELL_PATH "/bin/sh"

static bool processMPlayerLine(const std::string& s, size_t& pos, size_t& total)
{
  const std::string line = trim(s);
  if (line.empty() || line[0] != 'A')
    return 0;
  size_t state = 0;
  std::string str1, str2;
  for(std::string::size_type i = 0;i < line.length();i++)
    {
      if (line[i] == ' ')
	{
	  while (i < line.length() && line[i] == ' ')
	    i++;
	  i--;
	  state++;
	  continue;
	}
      if (state == 1)
	str1 += line[i];
      if (state == 4)
	str2 += line[i];
    }
  for(std::string::size_type i = 0;i < str1.length();i++)
    if (str1[i] == ',')
      str1[i] = '.';
  for(std::string::size_type i = 0;i < str2.length();i++)
    if (str2[i] == ',')
      str2[i] = '.';
  if (!checkTypeDouble(str1) || !checkTypeDouble(str2))
    return 0;
  const double fPos = parseAsDouble(str1);
  const double fTotal = parseAsDouble(str2);
  pos = (size_t)fPos;
  total = (size_t)fTotal;
  return 1;
}

void MPlayer::init()
{
  MUSITORIUS_SYS(pipe(m_inputPipe) == 0, "pipe()");
  MUSITORIUS_SYS(pipe(m_outputPipe) == 0, "pipe()");
  MUSITORIUS_SYS(pipe(m_errorPipe) == 0, "pipe()");
  logMsg(LOG_DEBUG, "MPlayer interface initialized");
}

bool MPlayer::play(const std::string& fileName, size_t volumeLevel)
{
  //FIXME:volume level is ignored by default;
  logMsg(LOG_DEBUG, "MPlayer interface received command to play \'%s\' with volume level %u", fileName.c_str(), volumeLevel);
  assert(m_pid == 0);
  std::string cmdLine;
  for(std::string::size_type i = 0;i < fileName.length();i++)
    if (fileName[i] == '\'')
      cmdLine += "\'\\\'\'"; else
      cmdLine += fileName[i];
  cmdLine = m_cmdLine + " \'" + cmdLine + "\'";
  logMsg(LOG_DEBUG, "Prepared MPlayer command line is \'%s\'", cmdLine.c_str());
  //FIXME:read all available data from input pipe;
  m_paused = 0;
  m_pid = fork();
  if (m_pid == (pid_t)-1)
    {
      logMsg(LOG_ERR, "Cannot create child process, %s", ERRNO_MSG);
      m_pid = 0;
      return 0;
    }
  if (m_pid == (pid_t)0)
    {
      close(m_inputPipe[1]);
      close(m_outputPipe[0]);
      close(m_errorPipe[0]);
      dup2(m_inputPipe[0], STDIN_FILENO);
            dup2(m_outputPipe[1], STDOUT_FILENO);
            dup2(m_errorPipe[1], STDERR_FILENO);
      if (execlp(SHELL_PATH, SHELL_PATH, "-c", cmdLine.c_str(), NULL) == -1)
	exit(EXIT_FAILURE);
    } // child process;
  logMsg(LOG_DEBUG, "MPlayer child process is executed with pid %d", m_pid);
  return 1;
}

void MPlayer::pause()
{
  assert(!m_paused);
  m_paused = 1;
  sendCommand("pause");
}

void MPlayer::resume()
{
  assert(m_paused);
  m_paused = 0;
  sendCommand("pause");
}

void MPlayer::stop()
{
  if (m_pid == 0)
    return;
  kill(m_pid, SIGINT);
  kill(m_pid, SIGKILL);
  const pid_t p = waitpid(m_pid, NULL, 0);
  if (p != m_pid)
    logMsg(LOG_WARNING, "waitpid() has returned unexpected pid after KILL signal: waiting %d but received %d", m_pid, p);
  m_pid = 0;
  if (p == -1)
    logMsg(LOG_ERR, "waitpid() failed after KILL signal: %s", ERRNO_MSG);
  m_paused = 0;
}

void MPlayer::jumpForward(size_t offset)
{
  assert(!m_paused);
  std::ostringstream ss;
  ss << "seek +" << offset;
  sendCommand(ss.str());
}

void MPlayer::jumpBackward(size_t offset)
{
  assert(!m_paused);
  std::ostringstream ss;
  ss << "seek -" << offset;
  sendCommand(ss.str());
}

void MPlayer::setVolume(size_t value)
{
  assert(!m_paused);
  //FIXME:
}

void MPlayer::onSigChld(pid_t pid)
{
  logMsg(LOG_DEBUG, "MPlayer interface is requested to handle new SIGCHLD with pid %d", pid);
  if (m_pid <= 0)
    {
      logMsg(LOG_WARNING, "Collected zombie with pid %d but MPlayer was not launched. What is it?", pid);
      return;
    }
  if (pid != m_pid)
    {
      logMsg(LOG_WARNING, "Collected zombie with pid %d but launched MPLayer has pid %d. What is it?", pid, m_pid);
      return;
    }
  m_pid = 0;
  m_paused = 0;
  m_callback.onPlayerStop(0);//We have no exit code analizing in this version;
}

int MPlayer::getOutputFd() const
{
  return m_outputPipe[0];
}

int MPlayer::getErrorFd() const
{
  return m_errorPipe[0];
}

void MPlayer::readOutputData()
{
  char buf[2048];
  ssize_t res = ::read(m_outputPipe[0], buf, sizeof(buf));
  if (res < 0)
    {
      logMsg(LOG_ERR, "Error reading data from MPlayer process: %s", ERRNO_MSG);
      return;
    }
  std::string line = m_outputChain;
  for(size_t i = 0;i < (size_t)res;i++)
    {
      if (buf[i] == '\r' || buf[i] == '\n')
	{
	  size_t sec = 0, total = 0;
	  if (processMPlayerLine(line, sec, total))
	    m_callback.onPlayerPosition(sec, total);
	  line.erase();
	} else 
	line += buf[i];
    }
  m_outputChain = line;
}

void MPlayer::readErrorData()
{
  char buf[2048];
  std::string s = m_errorChain;
  const ssize_t res = ::read(m_errorPipe[0], buf, sizeof(buf));
  if (res < 0)
    {
      logMsg(LOG_ERR, "Error reading error messages from MPlayer: %s", ERRNO_MSG);
      return;
    }
  for(size_t i = 0;i < (size_t)res;i++)
    {
      const char c = buf[i];
      if (c == '\r')
	continue;
      if (c == '\n')
	{
	  s = trim(s);
	  if (!s.empty())
	    logMsg(LOG_WARNING, "Received error message from MPlayer: %s", s.c_str());
	  s.erase();
	  continue;
	}
      s += c;
    } //for();
  m_errorChain = s;
}

void MPlayer::sendCommand(const std::string& command)
{
  //FIXME:check how many bytes are available for reading in input pipe;
  std::string s = command;
  s += "\n";
  const ssize_t res = writeBuffer(m_inputPipe[1], s.c_str(), s.length());
  if (res < 0)
    logMsg(LOG_ERR, "Could not send command to mplayer: %s (command is \'%s\')", ERRNO_MSG, command.c_str());
}

auto_ptr<AbstractPlayer> createMPlayerPlayer(const std::string& cmdLine, AbstractPlayerCallback& callback)
{
  return auto_ptr<AbstractPlayer>(new MPlayer(cmdLine, callback));
}
