
#include<assert.h>
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<sys/types.h>
#include<unistd.h>
#include<errno.h>
#include<sys/stat.h>
#include<fcntl.h>
#include<sys/socket.h>
#include<sys/un.h>
#include"musitoriusclient.h"

#define MUSITORIUS_SERVER_REPLY_OK 200

#define IO_BUF_SIZE 512
#define MAX_INPUT_LINE_LEN 1024

static unsigned int extractResultCode(char* line)
{
  unsigned int value = 0;
  size_t i;
  assert(line);
  for(i = 0;line[i] != '\0';i++)
    {
      char c = line[i];
      if (c < '0' || c > '9')
	return value;
      value *= 10;
      value += (c - '0');
    }
  return value;
}

void musitorius_string_list_init(musitorius_string_list_t* list)
{
  assert(list);
  list->first = NULL;
  list->last = NULL;
}

char musitorius_string_list_add(musitorius_string_list_t*list, char* value)
{
  musitorius_string_list_item_t* item = NULL;
  char* newValue = NULL;
  assert(list);
  assert(value);
  newValue = (char*)malloc(strlen(value) + 1);
  if (newValue == NULL)
    return MUSITORIUS_GENERAL_ERROR;
  strcpy(newValue, value);
  item = (musitorius_string_list_item_t*)malloc(sizeof(musitorius_string_list_item_t));
  if (item == NULL)
    {
      free(newValue);
      return MUSITORIUS_GENERAL_ERROR;
    }
  item->value = newValue;
  item->next = NULL;
  if (list->last != NULL)
    {
      list->last->next = item;
      list->last = item;
      return MUSITORIUS_GENERAL_OK;
    }
  list->first = item;
  list->last = item;
  return MUSITORIUS_GENERAL_OK;
}

void musitorius_string_list_destroy(musitorius_string_list_t* list)
{
  musitorius_string_list_item_t* item = NULL;
  assert(list);
  item = list->first;
  while(item)
    {
      musitorius_string_list_item_t* p = item->next;
      if (item->value != NULL)
	free(item->value);
      free(item);
      item = p;
    }
  list->first = NULL;
  list->last = NULL;
}

musitorius_connection_t musitorius_connect(char* path)
{
  /*FIXME:NULL path means the default socket*/
  struct sockaddr_un addr;
  int fd;
  assert(path);
  fd = socket(AF_UNIX, SOCK_STREAM, 0);
  if (fd == -1)
    return MUSITORIUS_BAD_CONNECTION;
  addr.sun_family = AF_UNIX;
  strncpy(addr.sun_path, path, sizeof(addr.sun_path));
  if (connect(fd, (struct sockaddr*)&addr, sizeof(struct sockaddr_un)) == -1)
    {
      close(fd);
      return MUSITORIUS_BAD_CONNECTION;
    }
  return fd;
}

void musitorius_close(musitorius_connection_t connection)
{
  if (connection == MUSITORIUS_BAD_CONNECTION)
    return;
  close(connection);
}

/*RWrites block of specified length and produces subsequent calls of write() in case of short write operation*/
static ssize_t writeblock(int fd, void* buf, size_t bufSize)
{
  char* b = (char*)buf;
  size_t c = 0;
  assert(buf);
  while(c < bufSize)
    {
      ssize_t res = write(fd, &b[c], bufSize - c);
      if (res == -1)
	return -1;
      assert(res >= 0);
      c += (size_t)res;
    } /*while();*/
  assert(c == bufSize);
  return (ssize_t)c;
}

/*Writes buffer of any length producing subsequent calls to write fixed length blocks*/
static ssize_t writebuf(int fd, void* buf, size_t bufSize)
{
  char* b = (char*)buf;
  size_t c = 0;
  assert(buf);
  while(c < bufSize)
    {
      size_t requiredSize = bufSize > c + IO_BUF_SIZE?IO_BUF_SIZE:(size_t)(bufSize - c);
      size_t res = writeblock(fd, &b[c], requiredSize);
      if (res == -1)
	return -1;
      assert(res == (ssize_t)requiredSize);
      c += (size_t)res;
    } /*while();*/
  assert(c == bufSize);
  return (ssize_t)c;
}

/*If there was some data after new line char it will be silently ignored>
 * This function returnes:
 *  something less zero - there was an error;
 *  0 - the end was reached (new line char or file descriptor is closed);
 *  1 - try again, there must be more data.
 */
static ssize_t read_until_new_line(int fd, char* buf, size_t* count)
{
  ssize_t res;
  size_t i;
  assert(buf);
  assert(count);
  res = read(fd, buf, *count);
  if (res < 0)
    return res;
  if (res == 0)
    {
      *count = 0;
      return 0;
    }
  *count = (size_t)res;
  for(i = 0;i < (size_t)res;i++)
    if (buf[i] == '\n')
      break;
  if (i == (size_t)res)
    return 1;
  *count = i;
  return 0;
}

static ssize_t read_line(int fd, char* line, size_t limit)
{
  ssize_t count = 0;
  assert(limit > 0);
  while(1)
    {
      char buf[IO_BUF_SIZE];
      size_t len = sizeof(buf);
      ssize_t res = read_until_new_line(fd, buf, &len);
      if (res < 0)
	return res;
      assert(len < sizeof(buf));
      if (len > 0)
	{
	  if (count + len > limit)
	    {
	      memcpy(&line[count], buf, limit - count);
	      count = limit;
	    } else 
	    {
	      memcpy(&line[count], buf, len);
	      count += len;
	    }
	}
      if (count == limit)
	line[count - 1] = '\0'; else
	if (res == 0)
	  line[count] = '\0';
      if (res == 0)
	return 0;
      if (count == limit)
	{
	  len = sizeof(buf);
	  while ((res = read_until_new_line(fd, buf, &len)) == 1)
	    len = sizeof(buf);
	  return res;
	}
    }
  assert(0);
  return -1;/*just to reduce warning messages*/
}

musitorius_result_t musitorius_task_single_file(musitorius_connection_t connection, char* path)
{
  char line[MAX_INPUT_LINE_LEN];
  ssize_t res;
  assert(connection != MUSITORIUS_BAD_CONNECTION);
  assert(path);
  res = writeblock(connection, "=TASK\n", 6);
  if (res < 0)
    return res;
  res = writeblock(connection, "+", 1);
  if (res < 0)
    return res;
  res = writebuf(connection, path, strlen(path));
  if (res < 0)
    return res;
  res = writeblock(connection, "\n-\n", 3);
  if (res < 0)
    return res;
  res = read_line(connection, line, sizeof(line));
  if (res < 0)
    return res;
  if (extractResultCode(line) == MUSITORIUS_SERVER_REPLY_OK)
    return MUSITORIUS_GENERAL_OK;
  return MUSITORIUS_GENERAL_ERROR;
}

musitorius_result_t musitorius_task(musitorius_connection_t connection, musitorius_string_list_t* list)
{
  char line[MAX_INPUT_LINE_LEN];
  musitorius_string_list_item_t* item = NULL;
  ssize_t res;
  assert(connection != MUSITORIUS_BAD_CONNECTION);
  assert(list);
  res = writeblock(connection, "=TASK\n", 6);
  if (res < 0)
    return res;
  item = list->first;
  while (item)
    {
      assert(item->value);
      res = writeblock(connection, "+", 1);
      if (res < 0)
	return res;
      res = writebuf(connection, item->value, strlen(item->value));
      if (res < 0)
	return res;
      res = writebuf(connection, "\n", 1);
      if (res < 0)
	return res;
      item = item->next;
    }
  res = writeblock(connection, "-\n", 2);
  if (res < 0)
    return res;
  res = read_line(connection, line, sizeof(line));
  if (res < 0)
    return res;
  if (extractResultCode(line) == MUSITORIUS_SERVER_REPLY_OK)
    return MUSITORIUS_GENERAL_OK;
  return MUSITORIUS_GENERAL_ERROR;
}

musitorius_result_t musitorius_play(musitorius_connection_t connection)
{
  char line[MAX_INPUT_LINE_LEN];
  ssize_t res;
  assert(connection != MUSITORIUS_BAD_CONNECTION);
  res = writeblock(connection, "=PLAY\n-\n", 8);
  if (res < 0)
    return res;
  res = read_line(connection, line, sizeof(line));
  if (res < 0)
    return res;
  if (extractResultCode(line) == MUSITORIUS_SERVER_REPLY_OK)
    return MUSITORIUS_GENERAL_OK;
  return MUSITORIUS_GENERAL_ERROR;
}

musitorius_result_t musitorius_stop(musitorius_connection_t connection)
{
  char line[MAX_INPUT_LINE_LEN];
  ssize_t res;
  assert(connection != MUSITORIUS_BAD_CONNECTION);
  res = writeblock(connection, "=STOP\n-\n", 8);
  if (res < 0)
    return res;
  res = read_line(connection, line, sizeof(line));
  if (res < 0)
    return res;
  if (extractResultCode(line) == MUSITORIUS_SERVER_REPLY_OK)
    return MUSITORIUS_GENERAL_OK;
  return MUSITORIUS_GENERAL_ERROR;
}

musitorius_result_t musitorius_next(musitorius_connection_t connection)
{
  char line[MAX_INPUT_LINE_LEN];
  ssize_t res;
  assert(connection != MUSITORIUS_BAD_CONNECTION);
  res = writeblock(connection, "=NEXT\n-\n", 8);
  if (res < 0)
    return res;
  res = read_line(connection, line, sizeof(line));
  if (res < 0)
    return res;
  if (extractResultCode(line) == MUSITORIUS_SERVER_REPLY_OK)
    return MUSITORIUS_GENERAL_OK;
  return MUSITORIUS_GENERAL_ERROR;
}

musitorius_result_t musitorius_prev(musitorius_connection_t connection)
{
  char line[MAX_INPUT_LINE_LEN];
  ssize_t res;
  assert(connection != MUSITORIUS_BAD_CONNECTION);
  res = writeblock(connection, "=PREV\n-\n", 8);
  if (res < 0)
    return res;
  res = read_line(connection, line, sizeof(line));
  if (res < 0)
    return res;
  if (extractResultCode(line) == MUSITORIUS_SERVER_REPLY_OK)
    return MUSITORIUS_GENERAL_OK;
  return MUSITORIUS_GENERAL_ERROR;
}

musitorius_result_t musitorius_forward(musitorius_connection_t connection, size_t step)
{
  char line[MAX_INPUT_LINE_LEN];
  ssize_t res;
  assert(connection != MUSITORIUS_BAD_CONNECTION);
  snprintf(line, sizeof(line), "=FORWARD %u\n-\n", step);
  res = writeblock(connection, line, strlen(line));
  if (res < 0)
    return res;
  res = read_line(connection, line, sizeof(line));
  if (res < 0)
    return res;
  if (extractResultCode(line) == MUSITORIUS_SERVER_REPLY_OK)
    return MUSITORIUS_GENERAL_OK;
  return MUSITORIUS_GENERAL_ERROR;
}

musitorius_result_t musitorius_backward(musitorius_connection_t connection, size_t step)
{
  char line[MAX_INPUT_LINE_LEN];
  ssize_t res;
  assert(connection != MUSITORIUS_BAD_CONNECTION);
  snprintf(line, sizeof(line), "=BACKWARD %u\n-\n", step);
  res = writeblock(connection, line, strlen(line));
  if (res < 0)
    return res;
  res = read_line(connection, line, sizeof(line));
  if (res < 0)
    return res;
  if (extractResultCode(line) == MUSITORIUS_SERVER_REPLY_OK)
    return MUSITORIUS_GENERAL_OK;
  return MUSITORIUS_GENERAL_ERROR;
}

